(****
   Denotational semantics for
   Biorthogonality, Step-Indexing & Compiler Correctness
   Nick Benton & Chung-Kil Hur
   April 2009
 ****)

Require Import utility.
Require Import Predom.
Require Import PredomAux.
Require Import typedlambda.

Set Implicit Arguments.
Unset Strict Implicit.

Open Scope O_scope.

(*==========================================================================
  Meaning of types and contexts
  ==========================================================================*)

Fixpoint SemType (t : Ty) {struct t} : cpo :=
match t with
 | Int => Discrete nat
 | Bool => Discrete bool
 | Arrow t1 t2 => (SemType t1) -C-> (DL (SemType t2))
 | t1 ** t2 => Dprod (SemType t1) (SemType t2)
end.

(* We define the semantics of an environment "from the right";
   this avoids need for SWAP in the semantics *)
Fixpoint SemEnv (e : Env) : cpo :=
match e with
 | nil => DOne
 | t :: ts => Dprod (SemEnv ts) (SemType t)
end.

(* Lookup in an environment *)
Fixpoint SemVar (E:Env) (t:Ty) (var:Var E t) {struct var} : SemEnv E -c> SemType t :=
  match var in Var E t return SemEnv E -c> SemType t with
  | ZVAR _ _ => SND
  | SVAR _ _ _ v => SemVar v @_ FST
  end.

(* A coercion to allow embedding values of type X into its lifted 
   discrete domain. DL_discrete is funny because we need to put in the  
   stuff the existing coercions are already adding or else the matching 
   will fail. The Id type operator is there because coercions are by 
   name. *)   

Definition Id (x : Type) := x.
Definition DL_discrete X := tord (tcpo (DL (Discrete X))).

Definition discrete_lift : forall X : Set, Id X -> DL_discrete X := 
  fun X x => eta (x : Discrete X).

Implicit Arguments discrete_lift [X].

Coercion discrete_lift : Id >-> DL_discrete.

(* The need to use the defined operator names -- especially Id --
   makes it seem like a marginal benefit. But maybe it's worth
   it as an example of how coercions work.

Definition top : DL_discrete unit := (tt : Id unit).
Definition top': DL (Discrete unit) := discrete_lift tt.

*) 
Definition indiscrete (X:Type) (x:X) : Discrete X := x.

Definition SimpleOpm (A:Type) (B:cpo) (op : A -> ((tord B) : Type)) : Discrete A -m> B.
intros.
  exists op.
  unfold monotonic.
  intros.
  simpl in *.
  rewrite H. auto.
Defined.

Definition SimpleOp (A:Set) (B:cpo) (op : A -> B) : Discrete A -c> B.
intros.
exists (SimpleOpm op).
unfold continuous.
intros.
assert (h 0 = lub h). auto.
rewrite <- H. simpl.
assert (op (h 0) <= (SimpleOpm op @ h) 0). auto.
eapply Ole_trans. apply H0.
apply le_lub.
Defined.

Definition SimpleOp2mm (A B C : Type) (op : A -> B -> C) : Discrete A -> Discrete B -m>  Discrete C.
intros.
exists (op X).
unfold monotonic.
intros.
simpl in *.
rewrite H. auto.
Defined.

Definition SimpleOp2c A B C (op:A -> B -> C) :
    Discrete A -> Discrete B -c> Discrete C.
intros.
exists (SimpleOp2mm op X).
simpl in *.
unfold continuous.
intros.
simpl. auto.
Defined.

Definition SimpleOp2m (A B C:Type) (op : A -> B -> C) :
    Discrete A -m> Discrete B -C-> Discrete C.
intros.
exists (SimpleOp2c op).
unfold monotonic.
intros.
simpl in *.
rewrite H. auto.
Defined.

Definition SimpleOp2 A B C (op:A -> B -> C) :
    Discrete A -c> Discrete B -C-> Discrete C.
intros.
exists (SimpleOp2m op).
unfold continuous ; intros ; simpl ; auto.
Defined.

Definition choosemm : forall (D : cpo), D -> D -> Discrete bool -m> D.
intros D X Y.
exists (fun (b:bool) => if b then X else Y).
unfold monotonic ; intros ; simpl in *; rewrite H ; auto.
Defined.

Definition choosec : forall (D : cpo), D -> D -> Discrete bool -c> D.
intros D X Y.
exists (choosemm X Y).
unfold continuous.
intros ; simpl in *.
assert ((if h 0 then X else Y) = ((choosemm X Y @ h) 0)).
simpl ; auto. rewrite H. apply le_lub.
Defined.

Definition choosem :  forall (D : cpo), D -> D -m> Discrete bool -C-> D.
intros.
exists (@choosec D X).
unfold monotonic.
intros.
simpl. intro.
destruct x0 ; auto.
Defined.

Definition choosecc :  forall (D : cpo), D -> D -c> Discrete bool -C-> D.
intros.
exists (@choosem D X).
unfold continuous.
intros.
simpl.
intro.
destruct x.
eapply Ole_trans.
assert (X <= ((Fcontit (Discrete bool) D @ (choosem X @ h) <_> true) 0)).
simpl ; auto. apply H.
apply le_lub.
refine (lub_le_compat _).
refine (fmon_le_intro _).
intros n.
auto.
Defined.

Definition choosecm : forall (D:cpo), D -m> D -C-> Discrete bool -C-> D.
intros.
exists (@choosecc D).
unfold monotonic.
intros.
simpl.
intros.
destruct x1. auto. auto.
Defined.

Definition choose : forall (D:cpo), D -c> D -C-> Discrete bool -C-> D.
intros.
exists (@choosecm D).
unfold continuous.
intros.
simpl. intros.
destruct x0.
refine (lub_le_compat (fmon_le_intro _)).
intros n. auto.

eapply Ole_trans.
assert (x <= ((Fcontit (Discrete bool) D @ ((Fcontit D (Discrete bool -C-> D) @ (choosecm D @ h)) <_> x)) <_> false) 0).
simpl ; auto. apply H.
apply le_lub.
Defined.

Canonical Structure Discrete.

(*==========================================================================
  Meaning of values and expressions
  ==========================================================================*)

Definition SemCTypeT E T := SemEnv E -C-> SemType T.

Definition SemCTypeL E T := SemEnv E -C-> DL (SemType T).

Fixpoint idxVar E T (x:Var E T) : nat :=
  match x with 
    | ZVAR _ _ => 0
    | SVAR _ _ _ x' => S (idxVar x')
  end.

Fixpoint projEnv E1 E2 : SemEnv (E1++E2) -C-> SemEnv E1 :=
  match E1 with
    | nil => @K (SemEnv E2) (SemEnv nil) tt
    | T::E1' => PROD_map (projEnv E1' E2) (@ID (SemType T))
  end.

Definition DVtoDC T (d:SemType T) : SemCTypeL nil T
  := (@K (SemEnv nil) (DL (SemType T)) (Val d)).

Definition ctoC (A B : cpo) (f : A -c> B) : A-C->B.
intros.
simpl.
exact f.
Defined.

Definition PROD_FUN (A B C : cpo): (C -C-> A) -C-> (C -C-> B) -C-> C -C-> Dprod A B.
intros.
eapply CURRY.
apply ctoC.
eapply fcont_comp.
Focus 2.
eapply uncurry.
apply PROD_MAP.
eapply CURRY.
apply ctoC.
eapply fcont_comp.
apply (@EV (Dprod C C)).
apply PROD_map.
apply ID.
apply duplicate.
Defined.

Lemma PROD_FUN_simpl: forall (A B C:cpo) fa fb, PROD_FUN A B C fa fb == PROD_fun fa fb.
intros.
apply (@fcont_eq_intro C (Dprod A B)); intros.
auto.
Qed.
Hint Resolve PROD_FUN_simpl.

Definition sem_precomp E' E A(op: E'-C->E) : (E-C->A) -C-> E'-C->A :=
  curry (@EV E A @_ PROD_map (@ID (E-C->A)) op).

Definition sem_op E A B (op:A-C->B) : (E-C->A) -C-> E-C->B := fcont_COMP E _ _ op.

Definition sem_op2 E A B C (op:A -C-> B -C-> C) :
 (E -C-> A) -C-> ((E -C-> B) -C-> (E -C-> C)).
intros.
eapply CURRY.
eapply CURRY.
eapply fcont_COMP.
apply (uncurry _ _ _ op).
apply ctoC.
eapply PROD_fun.
eapply fcont_comp.
eapply EV.
eapply PROD_fun.
eapply fcont_comp.
eapply FST.
eapply FST.
eapply SND.

eapply fcont_comp.
eapply EV.
eapply PROD_fun.
eapply fcont_comp.
eapply SND.
eapply FST.
eapply SND.
Defined.

Definition sem_op3 E A B C D (op:A -C-> B -C-> C -C-> D):
  (E-C->A)-C->(E-C->B)-C->(E-C->C)-C->(E-C->D).
intros.
eapply CURRY.
eapply CURRY.
eapply CURRY.
eapply fcont_COMP.
apply (uncurry _ _ _ (uncurry _ _ _ op)).
apply ctoC.
eapply PROD_fun.
eapply PROD_fun.

eapply fcont_comp.
eapply EV.
eapply PROD_fun.
eapply fcont_comp.
Focus 2.
apply FST.
eapply fcont_comp.
Focus 2.
apply FST.
apply FST.
eapply SND.

eapply fcont_comp.
eapply EV.
eapply PROD_fun.
eapply fcont_comp.
Focus 2.
apply FST.
eapply fcont_comp.
Focus 2.
apply FST.
apply SND.
eapply SND.

eapply fcont_comp.
eapply EV.
eapply PROD_fun.
eapply fcont_comp.
Focus 2.
apply FST.
apply SND.
eapply SND.
Defined.

Definition sem_fix E A B: (SemCTypeL (A::(A-->B)::E) B) -C-> SemCTypeT E (A-->B) :=
  curry(
    (@FIXP (SemType (A-->B)) _)
    @_ (uncurry _ _ _ ((CURRY _ _ _) @_ (CURRY _ _ _)))).

Definition sem_curry (E A B:cpo): (Dprod E A -C-> B) -C-> E -C-> (A -C-> B) := 
  CURRY E A B.

Definition sem_app (E A B:cpo) : (E-C->DL (A-C->DL B))-C->(E-C->DL A)-C->E-C->DL B :=
  curry( curry(
    (kleisli_cc _ _ (@EV _ _)) 
    @_ (uncurry _ _ _ (@Smash _ _))
    @_ (uncurry _ _ _ (uncurry _ _ _ (@PROD_FUN (DL(A-C->DL B)) (DL A)  E))))).

Fixpoint SemExp (env:Env) (t:Ty) (exp:Exp env t) {struct exp} : (SemEnv env -c> DL (SemType t)) :=
match exp with
| TOP op dt1 dt2 => sem_op _ eta (sem_op2 _ (SimpleOp2 (OpSem op)) (SemVal dt1) (SemVal dt2))
| TGT  dt1 dt2 => sem_op _ eta (sem_op2 _ (SimpleOp2 blt_nat) (SemVal dt2) (SemVal dt1))
| TAPP _ _ dt1 dt2 => sem_app _ _ _ (sem_op _ eta (SemVal dt1)) (sem_op _ eta (SemVal dt2))
| TVAL _ dt => sem_op _ eta (SemVal dt)
| TLET _ _ dt1 dt2 => sem_app _ _ _ (sem_op _ eta (sem_curry _ _ _ (SemExp dt2))) (SemExp dt1)
| TIF t dtc dt1 dt2 => sem_op3 _ (choose (DL (SemType t))) (SemExp dt1) (SemExp dt2) (SemVal dtc)
| TFST _ _ dtp => sem_op _ eta (sem_op _ FST (SemVal dtp))
| TSND _ _ dtp => sem_op _ eta (sem_op _ SND (SemVal dtp))
end

with SemVal (env:Env) (t:Ty) (val:Value env t) {struct val} : SemEnv env -c> SemType t :=
match val with
| TINT n => @K _ (Discrete nat) n
| TBOOL b => @K _ (Discrete bool) b
| TVAR _ i => SemVar i
| TLAM _ _ dtb => sem_curry _ _ _ (SemExp dtb)
| TFIX _ _ dtb => sem_fix _ _ _ (SemExp dtb)
| TPAIR _ _ dt1 dt2 => sem_op2 _ PAIR (SemVal dt1) (SemVal dt2)
end.

Lemma choosec_t_simpl: forall (D:cpo) (a b:D), choosec a b true = a.
intros D a b.
auto.
Qed.

Lemma choosec_f_simpl: forall (D:cpo) (a b:D), choosec a b false = b.
intros D a b.
auto.
Qed.

Lemma choose_t_simpl: forall D a b, choose D a b true = a.
auto.
Qed.
Lemma choose_f_simpl: forall D a b, choose D a b false = b.
auto.
Qed.

Add Parametric Morphism E t (e:Exp E t) : (SemExp e)
with signature (@Oeq (SemEnv E)) ==> (@Oeq (DL (SemType t)))
as SemExp_eq_compat.
intros e0 e1 eeq.
destruct eeq. split ; auto.
Qed.

Add Parametric Morphism E t (e:Value E t) : (SemVal e)
with signature (@Oeq (SemEnv E)) ==> (@Oeq (SemType t))
as SemVal_eq_compat.
intros e0 e1 eeq.
destruct eeq. split ; auto.
Qed.

